import numpy as np
from skcomponents.processing.processor import SignalProcessor


class GaussianNoise(SignalProcessor):

    def __init__(self, noise_level: float = 0.01):
        """
        Add gaussian noise to a signal

        Parameters
        ----------
        noise_level : float
            noise level in percent of signal
        """
        self.noise_level = noise_level

    def process_signal(self, signal: np.ndarray) -> np.ndarray:

        noise = signal * self.noise_level * np.random.standard_normal(signal.shape)
        return signal + noise


class PoissonCounting(SignalProcessor):

    def process_signal(self, signal: np.ndarray) -> np.ndarray:

        return np.random.poisson(signal)


class DarkCurrent(SignalProcessor):
    """
        Add dark signal to the measurements in electrons
    """

    def __init__(self, e_pixel_s):

        self.dark_rate = e_pixel_s

    def process_signal(self, signal: np.ndarray) -> np.ndarray:

        return signal + self.dark_rate


class ADU_noise(SignalProcessor):

    def __init__(self, specs, integration_time, do_noise=True, images_to_bin=1):

        """
            Calculate and add noise to the signal in ADU. This  model assumes that the system is linear so that the variance
            in the measured DN counts is given by:

            variance = (Signal + Dark)/G  + (Read_electrons/G)**2

            where Signal and Dark are measured in DN and G has unit of [electrons/DN]

        """

        self.do_noise = do_noise
        self.time = integration_time
        self.gain = specs.DetectorWell / 2 ** specs.DetectorBits
        self.dark = specs.DetectorDark * self.time
        self.read_noise_electrons = np.array(specs.DetectorReadNoise)

        self.images_to_bin = images_to_bin

    def process_signal(self, signal: np.ndarray) -> np.ndarray:

        var = (signal) / self.gain + (self.read_noise_electrons / self.gain) ** 2
        ADU_noise = np.random.normal(0, np.sqrt(var), np.shape(signal))

        if self.do_noise:
            return signal + ADU_noise
        if not self.do_noise:
            return signal

    def noise_estimate_ADU(self, signal):

        if self.do_noise:

            var = (signal) / self.gain + (self.read_noise_electrons / self.gain) ** 2
            ADU_noise = np.sqrt(var)

        if not self.do_noise:
            ADU_noise = np.zeros(signal.shape)

        return ADU_noise


class PRNU(SignalProcessor):

    def __init__(self, prnu, bias, do_systematics=True):
        """
            Includes  the given photo-response non-uniformity to the images
            Assumption:

            Signal out =  g*Signal_in + O

            where g is the PRNU variation (i.e. the variation in pixel response) and O is the offset

            prnu: an image of the photo-response non uniformity
            bias: offsets

        """
        self.prnu = prnu
        self.bias = bias
        self.do_systematics = do_systematics

    def process_signal(self, signal: np.ndarray) -> np.ndarray:

        if self.do_systematics:
            return signal * self.prnu
        if not self.do_systematics:
            return signal


class PoissonNoise(SignalProcessor):
    """
        Add Poisson noise to the images

    """
    def __init__(self, specs, do_noise=True, images_to_bin=1):

        self.specs = specs
        self.do_noise = do_noise

    def process_signal(self, signal: np.ndarray) -> np.ndarray:

        if self.do_noise:

            bins = self.images_to_bin
            dat = []
            if bins > 1:
                for x in np.arange(0, bins + 1):
                    dat.append(np.random.poisson(signal))
                signal_out = np.mean(dat, 0)
            if bins == 1:
                signal_out = np.random.poisson(signal)

            return signal_out

        if not self.do_noise:
            return signal


class ReadoutNoise(SignalProcessor):
    """
        Add ReadOut noise to the images

    """
    def __init__(self, specs, do_noise=True):

        self.specs = specs
        self.do_noise = do_noise
        self.read_noise_electrons = np.array(specs.DetectorReadNoise)

    def process_signal(self, signal: np.ndarray) -> np.ndarray:

        if self.do_noise:

            read_noise = np.random.normal(0, self.read_noise_electrons, np.shape(signal))

            return signal + read_noise

        if not self.do_noise:

            return signal


class RandomNoise(SignalProcessor):
    """
        Add Random Noise to the images

    """
    def __init__(self, specs, do_noise=True, images_to_bin=1):

        self.specs = specs
        self.do_noise = do_noise

    def process_signal(self, signal: np.ndarray) -> np.ndarray:

        if self.do_noise:
            noise = np.random.normal(0, np.sqrt(signal), np.shape(signal))

            return signal + noise

        if not self.do_noise:

            return signal


class DigitizationNoise(SignalProcessor):
    """
        Add additional digitization noise to the samples

    """
    def __init__(self, specs, do_noise=True, lsb=0.3):

        self.specs = specs
        self.do_noise = do_noise
        self.lsb = 2 * lsb

    def process_signal(self, signal: np.ndarray) -> np.ndarray:

        if self.do_noise:

            digitization_noise = np.random.normal(0, self.lsb, np.shape(signal))

            return signal + digitization_noise

        if not self.do_noise:

            return signal
